/* EOAccessAdditions.m created by lindberg on Mon 20-Dec-1999 */
/*-
 * Copyright (c) 2001 Doug McClure
 *
 * Original code from eogenerator Copyright (c) 2001 Carl Lindberg and Mike Gentry
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#import "EOAccessAdditions.h"
#import "FoundationAdditions.h"

#ifndef EOF2_ONLY
#import <EOModeler/EOModelExtensions.h>

// <EOJavaClient/EOAccessAdditions.h> isn't always there (the ObjC framework
// does not come with WO 4.5.1 on OS X or with WO 5), so just declare the
// methods here.  Except with WO 4.0, the methods are actually implemented
// in the EOAccess framework, so this is OK.
@interface EOEntity (JavaClientExtensions)
- (NSString *)clientClassName;
- (NSArray *)clientClassProperties;
- (NSArray *)clientClassPropertyAttributeNames;
- (NSArray *)clientClassPropertyToOneRelationshipNames;
- (NSArray *)clientClassPropertyToManyRelationshipNames;
@end
#endif

@implementation EOModel (EOGenerator)

/*"
 * Tries to find the framework name. First looks to see if a value for
 * "EOGeneratorFrameworkName" exists in the model's -userInfo dictionary, and if
 * not, looks at the model's -path to find the framework the .eomodeld is in.
 * Both PB.project files and path components ending in ".framework" are looked
 * for, so it should work for .eomodels insided installed frameworks or in a
 * development tree. If the framework name could not be found, returns nil.
 * Several other custom methods use this information to generate \#import
 * statements better. If all your EO classes are in the same framework, this
 * doesn't make much difference, but if there are EOModels in several different
 * frameworks with interrelationships, these methods can make life a lot easier.
"*/
- (NSString *)frameworkName
{
    NSDictionary  *userInfo = [self userInfo];
    NSFileManager *manager  = [NSFileManager defaultManager];
    NSString      *name;

    name = [userInfo objectForKey:@"EOGeneratorFrameworkName"];

    if (name == nil)
    {
        NSString *fullPath = [self path];
        NSMutableDictionary *mutableInfo = [[userInfo mutableCopy] autorelease];

        if (mutableInfo == nil) mutableInfo = [NSMutableDictionary dictionary];

        if (fullPath != nil && ![fullPath isAbsolutePath])
            fullPath = [[manager currentDirectoryPath] stringByAppendingPathComponent:fullPath];

        while ([fullPath isAbsolutePath])
        {
            NSString *pbPath;

            fullPath = [fullPath stringByDeletingLastPathComponent];
            pbPath = [fullPath stringByAppendingPathComponent:@"PB.project"];

            if ([manager regularFileExistsAtPath:pbPath])
            {
                NSDictionary *project = [NSDictionary dictionaryWithContentsOfFile:pbPath];

                if ([[project objectForKey:@"PROJECTTYPE"] hasSuffix:@"Framework"])
                {
                    name = [project objectForKey:@"PROJECTNAME"];
                    break;
                }
            }
            else if ([[fullPath pathExtension] isEqual:@"framework"])
            {
                name = [[fullPath lastPathComponent] stringByDeletingPathExtension];
                break;
            }
        }

        [mutableInfo setObject:name? name : @"" forKey:@"EOGeneratorFrameworkName"];
        [self setUserInfo:mutableInfo];
    }

    return [name length] > 0 ? name : nil;
}

- (NSArray *)entityNamesMatchingWildcard:(NSString *)wildcard
{
    NSMutableArray *names = [NSMutableArray array];
    NSArray *entities = [self entities];
    int i, count = [entities count];

    for (i=0; i<count; i++)
    {
        NSString *entityName = [[entities objectAtIndex:i] name];
        if ([entityName isLike:wildcard])
            [names addObject:entityName];
    }

    return names;
}

- (BOOL)hasParentInheritance
{
    NSEnumerator *enumerator = [[self entities] objectEnumerator];
    EOEntity *entity;

    while ( entity = [enumerator nextObject] ) {
        if ( [entity parentEntity] != nil ) {
            return YES;
        }
    }

    return NO;
}

- (BOOL)hasChildInheritance
{
    NSEnumerator *enumerator = [[self entities] objectEnumerator];
    EOEntity *entity;

    while ( entity = [enumerator nextObject] ) {
        if ( [[entity subEntities] count] > 0 ) {
            return YES;
        }
    }

    return NO;
}

- (BOOL)hasAbstractEntity
{
    NSEnumerator *enumerator = [[self entities] objectEnumerator];
    EOEntity *entity;

    while ( entity = [enumerator nextObject] ) {
        if ( [entity isAbstractEntity] ) {
            return YES;
        }
    }

    return NO;
}

- (BOOL)hasReadOnlyEntity
{
    NSEnumerator *enumerator = [[self entities] objectEnumerator];
    EOEntity *entity;

    while ( entity = [enumerator nextObject] ) {
        if ( [entity isReadOnly] ) {
            return YES;
        }
    }

    return NO;
}

- (NSArray *)concreteEntityNames
{
    NSEnumerator *enumerator = [[self concreteEntities] objectEnumerator];
    NSMutableArray *entityNames = [NSMutableArray array];
    EOEntity *entity;

    while ( entity = [enumerator nextObject] )
        [entityNames addObject:[entity name]];

    return entityNames;
}

- (NSArray *)concreteEntities
{
    NSEnumerator *enumerator = [[self entities] objectEnumerator];
    NSMutableArray *entities = [NSMutableArray array];
    EOEntity *entity;

    while ( entity = [enumerator nextObject] )
        if ( ![[entity classNameWithoutPackage] isEqualToString:@"EOGenericRecord"] )
            [entities addObject:entity];

    return entities;
}

- (NSArray *)genericEntityNames
{
    NSEnumerator *enumerator = [[self genericEntities] objectEnumerator];
    NSMutableArray *entityNames = [NSMutableArray array];
    EOEntity *entity;

    while ( entity = [enumerator nextObject] )
        [entityNames addObject:[entity name]];

    return entityNames;
}

- (NSArray *)genericEntities
{
    NSEnumerator *enumerator = [[self entities] objectEnumerator];
    NSMutableArray *entities = [NSMutableArray array];
    EOEntity *entity;

    while ( entity = [enumerator nextObject] )
        if ( [[entity classNameWithoutPackage] isEqualToString:@"EOGenericRecord"] )
            [entities addObject:entity];

    return entities;
}

@end


@implementation EOEntity (EOGenerator)

- (BOOL)hasParentEntity
{
    return ([self parentEntity] != nil);
}

- (NSString *)frameworkName
{
    return [[self model] frameworkName];
}

/*"
 * Returns <Framework/ClassName.h> if inside a framework, "ClassName.h"
 * otherwise.
"*/
- (NSString *)objcImportString
{
    return [self objcImportStringInRelationToEntity:nil];
}

/*"
 * If sourceEntity is nil, returns <Framework/ClassName.h> if inside a
 * framework, "ClassName.h" otherwise. If sourceEntity is non-nil, the
 * <Framework/ClassName.h> version will only be returned if the receiver's
 * framework and sourceEntity's framework are different.  This depends on
 * the -#frameworkName method doing the right thing.
"*/
- (NSString *)objcImportStringInRelationToEntity:(EOEntity *)sourceEntity
{
    NSString *myClass = [self className];
    NSString *myFramework = [self frameworkName];
    NSString *sourceFramework = [sourceEntity frameworkName];

    if ([myClass hasSuffix:@"EOGenericRecord"]) return nil;

    if (myFramework != nil &&
        (sourceEntity == nil || ![sourceFramework isEqualToString:myFramework]))
    {
        return [NSString stringWithFormat:@"<%@/%@.h>", myFramework, myClass];
    }
    else
    {
        return [NSString stringWithFormat:@"\"%@.h\"", myClass];
    }
}

/*"
 * If there is no parent entity, returns nil. Otherwise, uses
 * -#objcImportStringInRelationToEntity: to determine the string needed to
 * \#import the parent class. I'm not sure if EOModeler lets you have the parent
 * entity in an external .eomodeld, but we'll provide this to be safe.
"*/
- (NSString *)parentObjCImportString
{
    return [[self parentEntity] objcImportStringInRelationToEntity:self];
}

/*"
 * If -#parentObjCImportString is nil, returns an empty array, otherwise returns
 * a single-element array with the parent import string.
"*/
- (NSArray *)arrayWithParentObjCImportStringIfNeeded
{
    NSString *import = [self parentObjCImportString];

    if (import != nil)
        return [NSArray arrayWithObject:import];
    else
        return [NSArray array];
}

/*"
 * A parallel for -#referencedClassNames for dealing with import strings. The
 * array will contain a uniqued list of \#import strings for all related
 * entities.
"*/
- (NSArray *)referencedObjCImportStrings
{
    NSArray        *relationships = [self relationships];
    int            i, count = [relationships count];
    NSMutableArray *refImports = [NSMutableArray arrayWithCapacity:count];

    for (i=0; i<count; i++)
    {
        EORelationship *relationship = [relationships objectAtIndex:i];
        EOEntity       *destEntity   = [relationship destinationEntity];

        if (![[destEntity className] hasSuffix:@"EOGenericRecord"])
        {
            NSString *import = [relationship objcImportString];
            if (import && ![refImports containsObject:import])
                [refImports addObject:import];
        }
    }

    return [refImports sortedArrayUsingSelector:@selector(compare:)];
}

- (NSDictionary *)storedProcedures {
    return _storedProcedures;
}

- (EOStoredProcedure *)storedProcedureForFetchAll {
    return [self storedProcedureForOperation:EOFetchAllProcedureOperation];
}

- (EOStoredProcedure *)storedProcedureForFetchWithPrimaryKey {
    return [self storedProcedureForOperation:EOFetchWithPrimaryKeyProcedureOperation];
}

- (EOStoredProcedure *)storedProcedureForInsert {
    return [self storedProcedureForOperation:EOInsertProcedureOperation];
}

- (EOStoredProcedure *)storedProcedureForDelete {
    return [self storedProcedureForOperation:EODeleteProcedureOperation];
}

- (EOStoredProcedure *)storedProcedureForPrimaryKeyGeneration {
    return [self storedProcedureForOperation:EONextPrimaryKeyProcedureOperation];
}

- (BOOL)hasReadFormatOnAnAttribute {
    NSEnumerator *enumerator = [[self attributes] objectEnumerator];
    EOAttribute *anAttr;

    while ( anAttr = [enumerator nextObject] ) {
        if ( [anAttr readFormat] != nil ) {
            return YES;
        }
    }

    return NO;
}

- (BOOL)hasWriteFormatOnAnAttribute {
    NSEnumerator *enumerator = [[self attributes] objectEnumerator];
    EOAttribute *anAttr;

    while ( anAttr = [enumerator nextObject] ) {
        if ( [anAttr writeFormat] != nil ) {
            return YES;
        }
    }

    return NO;
}

- (BOOL)hasPrototypeOnAnAttribute {
    NSEnumerator *enumerator = [[self attributes] objectEnumerator];
    EOAttribute *anAttr;

    while ( anAttr = [enumerator nextObject] ) {
        if ( [anAttr prototype] != nil ) {
            return YES;
        }
    }

    return NO;
}

- (BOOL)hasDefinitionOnAnAttribute {
    NSEnumerator *enumerator = [[self attributes] objectEnumerator];
    EOAttribute *anAttr;

    while ( anAttr = [enumerator nextObject] ) {
        if ( [anAttr definition] != nil ) {
            return YES;
        }
    }

    return NO;
}

- (BOOL)hasValueTypeOnAnAttribute {
    NSEnumerator *enumerator = [[self attributes] objectEnumerator];
    EOAttribute *anAttr;

    while ( anAttr = [enumerator nextObject] ) {
        if ( [anAttr valueType] != nil ) {
            return YES;
        }
    }

    return NO;
}


- (BOOL)hasDefinitionOnARelationship {
    NSEnumerator *enumerator = [[self relationships] objectEnumerator];
    EORelationship *aRelationship;

    while ( aRelationship = [enumerator nextObject] ) {
        if ( [aRelationship definition] != nil ) {
            return YES;
        }
    }

    return NO;
}

#ifndef EOF2_ONLY
/* 
 * Implementations for some JavaClient-specific methods that EOModeler gets
 * because they are in the EOJavaClientExtensions.EOMBundle. We could possibly
 * load the bundle, but this is easier.
 */

- (NSArray *)clientClassAttributes
{
    NSArray *names = [self clientClassPropertyAttributeNames];
    int     i, count = [names count];
    NSMutableArray *attributes = [NSMutableArray arrayWithCapacity:count];

    for (i=0; i<count; i++)
        [attributes addObject:[self attributeNamed:[names objectAtIndex:i]]];

    return attributes;
}

- (NSArray *)clientClassToOneRelationships
{
    NSArray *names = [self clientClassPropertyToOneRelationshipNames];
    int     i, count = [names count];
    NSMutableArray *relationships = [NSMutableArray arrayWithCapacity:count];

    for (i=0; i<count; i++)
        [relationships addObject:[self relationshipNamed:[names objectAtIndex:i]]];

    return relationships;
}

- (NSArray *)clientClassToManyRelationships
{
    NSArray *names = [self clientClassPropertyToManyRelationshipNames];
    int     i, count = [names count];
    NSMutableArray *relationships = [NSMutableArray arrayWithCapacity:count];

    for (i=0; i<count; i++)
        [relationships addObject:[self relationshipNamed:[names objectAtIndex:i]]];

    return relationships;
}

// These next two aren't used, but the bundle implements them, so we might as
// well too in case someone wants to use them.
- (NSArray *)clientClassScalarAttributes
{
    NSArray *attributes = [self clientClassAttributes];
    int     i, count = [attributes count];
    NSMutableArray *scalars = [NSMutableArray arrayWithCapacity:count];

    for (i=0; i<count; i++)
    {
        EOAttribute *attrib = [attributes objectAtIndex:i];
        if ([attrib isScalar])
            [scalars addObject:attrib];
    }

    return scalars;
}

- (NSArray *)clientClassNonScalarAttributes
{
    NSArray *attributes = [self clientClassAttributes];
    int     i, count = [attributes count];
    NSMutableArray *nonscalars = [NSMutableArray arrayWithCapacity:count];

    for (i=0; i<count; i++)
    {
        EOAttribute *attrib = [attributes objectAtIndex:i];
        if (![attrib isScalar])
            [nonscalars addObject:attrib];
    }

    return nonscalars;
}

#endif

#ifdef EOF2_ONLY
/* Define for use by EOF2 systems */
- (NSString *)classNameWithoutPackage
{
    NSString *className = [self className];
    NSRange  dotRange = [className rangeOfString:@"." options:NSBackwardsSearch];

    if (dotRange.length > 0)
        className = [className substringFromIndex:NSMaxRange(dotRange)];

    return className;
}

- (NSArray *)fetchSpecificationNames
{
    return [NSArray array];
}
- (EOFetchSpecification *)fetchSpecificationNamed:(NSString *)name
{
    return nil;
}
- (NSArray *)clientClassProperties
{
    return [NSArray array];
}

#endif

- (NSArray *)fetchSpecifications
{
    NSMutableArray *fetchSpecifications = [NSMutableArray arrayWithCapacity:[[self fetchSpecificationNames] count]];
    NSEnumerator *fetchNameEnumerator = [[self fetchSpecificationNames] objectEnumerator];
    NSString *fetchName;

    while ( fetchName = [fetchNameEnumerator nextObject] ) {
        [fetchSpecifications addObject:[NSDictionary dictionaryWithObjectsAndKeys:fetchName, @"name",
            [self fetchSpecificationNamed:fetchName], @"spec", nil]];
    }

    return fetchSpecifications;
}

@end


@implementation EOAttribute (EOGenerator)

- (BOOL)isPrimaryKey
{
    EOEntity *parentEntity = [self entity];
    return [[parentEntity primaryKeyAttributes] containsObject:self];
}

- (BOOL)isClassProperty
{
    EOEntity *parentEntity = [self entity];
    return [[parentEntity classProperties] containsObject:self];
}

- (BOOL)isClientClassProperty
{
    EOEntity *parentEntity = [self entity];
    return [[parentEntity clientClassProperties] containsObject:[self name]];
}

- (BOOL)isSnapshotProperty
{
    EOEntity *parentEntity = [self entity];
    return [[parentEntity attributesUsedForLocking] containsObject:self];
}

- (NSString *)fullExternalType
{
    NSString *type = [self externalType];
    unsigned precision = [self precision];
    unsigned width     = [self width];
    int      scale     = [self scale];

    if (precision && scale)
        return [NSString stringWithFormat:@"%@(%d,%d)", type, precision, scale];
    if (width || precision)
        return [NSString stringWithFormat:@"%@(%d)", type, width?width:precision];

    return type;
}

#ifdef EOF2_ONLY
// Define for use by EOF2 systems
- (id)prototype
{
    return nil;
}
#endif

@end


@implementation EORelationship (EOGenerator)

- (NSString *)objcImportString
{
    return [[self destinationEntity] objcImportStringInRelationToEntity:[self entity]];
}

- (BOOL)isClassProperty
{
    EOEntity *parentEntity = [self entity];
    return [[parentEntity classProperties] containsObject:self];
}

- (BOOL)isClientClassProperty
{
    EOEntity *parentEntity = [self entity];
    return [[parentEntity clientClassProperties] containsObject:self];
}

@end
