/* EOTemplateGenerator.m created by mgentry on Mon 14-Jun-1999 */
/*-
 * Copyright (c) 2001 Doug McClure
 *
 * Original code from eogenerator Copyright (c) 2001 Carl Lindberg and Mike Gentry
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#import "EOTemplateGenerator.h"
#import <EOAccess/EOModel.h>
#import <EOModeler/EOModelExtensions.h>
#import <MiscMerge.subproj/MiscMergeTemplate.h>
#import <MiscMerge.subproj/MiscMergeCommandBlock.h>
#import <MiscMerge.subproj/MiscMergeEngine.h>
#import <MiscFoundation.subproj/NSString+MiscAdditions.h>
#import "EOAccessAdditions.h"
#import "FoundationAdditions.h"
#import "SystemType.h"


#define DefaultObjCGroupTemplate @"ObjcGroup.doctemplate"
#define DefaultJavaGroupTemplate @"JavaGroup.doctemplate"
#define DefaultGroupFile @"index.html"

#define DefaultObjCModelTemplate @"ObjcModel.doctemplate"
#define DefaultJavaModelTemplate @"JavaModel.doctemplate"
#define DefaultModelExt @"html"

#define DefaultObjCEntityTemplate @"ObjcEntity.doctemplate"
#define DefaultJavaEntityTemplate @"JavaEntity.doctemplate"
#define DefaultEntityExt @"html"

#define DefaultIndexFile @"index.html"

/* Writes to stderr */
void ErrVPrintf(NSString *format, va_list arguments)
{
    NSString *logString = [[NSString alloc] initWithFormat:format arguments:arguments];
    fwrite([logString cString], 1, [logString cStringLength], stderr);
    if (![logString hasSuffix:@"\n"]) fputc('\n', stdout);
    [logString release];
}

void ErrPrintf(NSString *format, ...)
{
    va_list arguments;

    va_start(arguments, format);
    ErrVPrintf(format, arguments);
    va_end(arguments);
}

void VPrintf(NSString *format, va_list arguments)
{
    NSString *logString = [[NSString alloc] initWithFormat:format arguments:arguments];
    fwrite([logString cString], 1, [logString cStringLength], stdout);
    [logString release];
}

void Printf(NSString *format, ...)
{
    va_list arguments;

    va_start(arguments, format);
    VPrintf(format, arguments);
    va_end(arguments);
}


@implementation EOTemplateGenerator

- init
{
    [super init];

    /* Set our own default group. The +globalModelGroup does some command line
    parsing, which certain parameter values can mess up causing exceptions to
    be thrown. We don't want any of the self-discovery of models anyways, so
    just avoid it. */
    [EOModelGroup setDefaultGroup:[[EOModelGroup alloc] init]];

    verboseMode = NO;
    forceOverwrites = NO;
    docDestination = @".";
    docTemplate = nil;
    docExtension = nil;
    docFile = nil;

    models = [[NSMutableArray alloc] init];
    searchPaths = [[NSMutableArray alloc] init];
    entityNames = [[NSMutableArray alloc] init];
    entityWildcards = [[NSMutableArray alloc] init];
    userVariables = [[NSMutableDictionary alloc] init];

    delimStart = @"[$";
    delimEnd = @"$]";

    docMode = ArgumentUnknownOption;
    listEntityTypes = ArgumentUnknownOption;

    return self;
}

- (void)displayHelpAndExit:(int)exitStatus
{
    NSString *appName = [[NSProcessInfo processInfo] processName];

    Printf(@"Usage: %@ must be used in one of four ways:\n", appName);
    Printf(@" %@ -groupdoc] -models model.eomodeld [model2.eomodeld ...] \n", appName);
    Printf(@" %@ -modeldoc -models model.eomodeld [model2.eomodeld ...]\n", appName);
    Printf(@" %@ -entitydoc -models model.eomodeld [model2.eomodeld ...] [entities ...]\n", appName);
    Printf(@" %@ -entitylist -models model.eomodeld [model2.eomodeld ...] [entities ...]\n", appName);
    Printf(@"%s",
           "\n"
           "Options:\n"
           " -models model [model ...]     Output entities from <model>\n"
           " -refmodels model [model ...]  Load model, do not generate code\n"
           "\n"
           " -concreteonly                 Only list/document entities that are not EOGenericRecords\n"
           "                               (only applies to -entitydoc and -entitylist)\n"
           " -genericonly                  Only list/document entities that are EOGenericRecords\n"
           "                               (only applies to -entitydoc and -entitylist)\n"
           "\n"
           " -listformat <string>          Use <string> as the format for -entitylist\n"
           "                               Use %%m for the model name, and %%e for the entity name\n"
           " -listseparator <string>       Use <string> as the separator between lines for -entitylist\n"
           "\n"
           " -templatedir <dir>            Add <dir> to template search path\n"
           " -template <file>              Use <file> as the template to generate\n"
           "\n"
           " -destination <dir>            Output directory for generated files\n"
           " -extension <ext>              Output file with <ext> on the end\n"
           " -file <file>                  Output file as <file>.<ext> in destination\n"
           " -indexfile                    Output file will be named \"index.html\"\n"
           "\n"
           " -delim <string>               Use <string> as start/end delimiters\n"
           " -startdelim <string>          Use <string> as start delimiters\n"
           " -enddelim <string>            Use <string> as end delimiters\n"
           "\n"
           " -force                        Force overwrite of read-only files\n"
           " -java                         Tells system the models should be are being\n"
           " -define-<key> <value>         Add <key> to runtime variable list\n"
           "\n"
           " -version                      Print version number and exit\n"
           " -help                         This help\n"
           " -verbose                      Verbose output\n"
           );
    
    exit(exitStatus);
}

- (void)addModelFromPath:(NSString *)path shouldProcess:(BOOL)shouldProcess
{
    EOModel *model = nil;

    NS_DURING
        model = [[EOModel alloc] initWithContentsOfFile:path];
    NS_HANDLER
        ErrPrintf(@"Error loading eomodel file '%@': %@", path, [localException reason]);
        return;
    NS_ENDHANDLER

    if ([[EOModelGroup defaultGroup] modelNamed:[model name]] == nil)
        [[EOModelGroup defaultGroup] addModel:model];

    if (model && shouldProcess)
        [models addObject:model];

    [model release];
}

- (void)addDefaultSearchPaths
{
    NSArray *libraryPaths;
    int i;

#ifdef OPENSTEP_ONLY
    libraryPaths = NSStandardLibraryPaths();
#else
    libraryPaths = NSSearchPathForDirectoriesInDomains(
                                                       NSAllLibrariesDirectory,
                                                       NSUserDomainMask|NSLocalDomainMask|NSNetworkDomainMask,
                                                       YES);
#endif

    for (i=0; i<[libraryPaths count]; i++) {
        NSString *libPath = [libraryPaths objectAtIndex:i];
        [searchPaths addObject:[libPath stringByAppendingPathComponent:@"EOReporter"]];
    }

    [searchPaths addObject:[[NSBundle mainBundle] bundlePath]];

    if (verboseMode)
        ErrPrintf(@"Search path is:\n%@", searchPaths);
}

- (NSArray *)searchPaths {
    return searchPaths;
}


- (int)argumentType:(NSString *)argument
{
    if ( [@"-version" hasPrefix:argument] )		return ArgumentVersion;
    else if ( [@"-help" hasPrefix:argument] )		return ArgumentHelp;
    else if ( [@"-verbose" hasPrefix:argument] )	return ArgumentVerbose;

    else if ( [@"-delim" hasPrefix:argument] )		return ArgumentDelim;
    else if ( [@"-startdelim" hasPrefix:argument] )	return ArgumentStartDelim;
    else if ( [@"-enddelim" hasPrefix:argument] )	return ArgumentEndDelim;

    else if ( [@"-force" hasPrefix:argument] )		return ArgumentForce;

    else if ( [@"-java" hasPrefix:argument] )		return ArgumentJava;

    else if ( [@"-models" hasPrefix:argument] )		return ArgumentModel;
    else if ( [@"-refmodels" hasPrefix:argument] )	return ArgumentRefModel;

    else if ( [@"-template" hasPrefix:argument] )	return ArgumentTemplate;
    else if ( [@"-templatedir" hasPrefix:argument] )	return ArgumentTemplateDir;

    else if ( [@"-destination" hasPrefix:argument] )	return ArgumentDestination;
    else if ( [@"-extension" hasPrefix:argument] )	return ArgumentExtension;
    else if ( [@"-file" hasPrefix:argument] )		return ArgumentFile;
    else if ( [@"-indexfile" hasPrefix:argument] )	return ArgumentIndexFile;
    else if ( [@"-groupdoc" hasPrefix:argument] )	return ArgumentGroupDoc;
    else if ( [@"-modeldoc" hasPrefix:argument] )	return ArgumentModelDoc;
    else if ( [@"-entitydoc" hasPrefix:argument] )	return ArgumentEntityDoc;
    else if ( [@"-entitylist" hasPrefix:argument] )	return ArgumentEntityList;
    else if ( [@"-listformat" hasPrefix:argument] )	return ArgumentListFormat;
    else if ( [@"-listseparator" hasPrefix:argument] )	return ArgumentListSeparator;
    
    else if ( [@"-genericonly" hasPrefix:argument] )	return ArgumentGenericEntityOnly;
    else if ( [@"-concreteonly" hasPrefix:argument] )	return ArgumentConcreteEntityOnly;

    else if ( [argument hasPrefix:@"-define-"] )	return ArgumentDefine;

    else if ( [argument hasPrefix:@"-"] )		return ArgumentInvalidOption;
    else return ArgumentUnknownOption;
}

- (void)processCommandLineArguments
{
    NSEnumerator *argEnum = [[[NSProcessInfo processInfo] arguments] objectEnumerator];
    NSString *currArg;
    int argType;

    [argEnum nextObject]; //skip argv[0]

    while ( currArg = [argEnum nextObject] ) {
TopArgProcess:
        argType = [self argumentType:currArg];

        switch ( argType ) {
            case ArgumentVersion:
                ErrPrintf(@"EOReporter version 1.0");
                exit(0);
                break;

            case ArgumentHelp:
                [self displayHelpAndExit:0];
                break;

            case ArgumentVerbose:
                verboseMode = YES;
                break;


            case ArgumentDelim:
                if ( (currArg = [argEnum nextObject]) && [currArg length] > 0 ) {
                    delimStart = [currArg retain];
                    delimEnd = [currArg retain];
                }
                else
                    ErrPrintf(@"Warning: -delim option specify an non-zero length delimiter");
                break;

            case ArgumentStartDelim:
                if ( (currArg = [argEnum nextObject]) && [currArg length] > 0 )
                    delimStart = [currArg retain];
                else
                    ErrPrintf(@"Warning: -startdelim option specify an non-zero length delimiter");
                break;

            case ArgumentEndDelim:
                if ( (currArg = [argEnum nextObject]) && [currArg length] > 0 )
                    delimEnd = [currArg retain];
                else
                    ErrPrintf(@"Warning: -enddelim option specify an non-zero length delimiter");
                break;


            case ArgumentForce:
                forceOverwrites = YES;
                break;

            case ArgumentJava:
                javaMode = YES;
                break;


            case ArgumentModel:
                while ( (currArg = [argEnum nextObject]) && [[currArg pathExtension] hasPrefix:@"eomodel"] )
                    [self addModelFromPath:currArg shouldProcess:YES];
                if ( currArg != nil )
                    goto TopArgProcess;
                    break;

            case ArgumentRefModel:
                while ( (currArg = [argEnum nextObject]) && [[currArg pathExtension] hasPrefix:@"eomodel"] )
                    [self addModelFromPath:currArg shouldProcess:NO];
                if ( currArg != nil )
                    goto TopArgProcess;
                    break;


            case ArgumentTemplateDir:
                if ( currArg = [argEnum nextObject] )
                    if ( ![[NSFileManager defaultManager] directoryExistsAtPath:currArg] )
                        ErrPrintf(@"Warning: Template search directory %@ does not exist", currArg);
                    else
                        [searchPaths addObject:currArg];
                else
                    ErrPrintf(@"Warning: template search directory must be a non-zero length path", currArg);
                break;

            case ArgumentTemplate:
                if ( (currArg = [argEnum nextObject]) && [currArg length] > 0 )
                    docTemplate = [currArg retain];
                else
                    ErrPrintf(@"Warning: template must be a non-zero length name");
                break;


            case ArgumentDestination:
                if ( (currArg = [argEnum nextObject]) && [currArg length] > 0 )
                    docDestination = [currArg retain];
                else
                    ErrPrintf(@"Warning: destination must be a non-zero length path");
                break;

            case ArgumentExtension:
                if ( (currArg = [argEnum nextObject]) && [currArg length] > 0 )
                    docExtension = [currArg retain];
                else
                    ErrPrintf(@"Warning: extension must be a non-zero length name", currArg);
                break;

            case ArgumentFile:
                if ( (currArg = [argEnum nextObject]) && [currArg length] > 0 )
                    docFile = [currArg retain];
                else
                    ErrPrintf(@"Warning: file must be a non-zero length name", currArg);
                break;

            case ArgumentIndexFile:
                indexFile = YES;
                if ( docFile == nil ) docFile = DefaultIndexFile;
                break;


            case ArgumentGroupDoc:
            case ArgumentModelDoc:
            case ArgumentEntityDoc:
            case ArgumentEntityList:
                if ( docMode == ArgumentUnknownOption )
                    docMode = argType;
                else {
                    ErrPrintf(@"Error: Can only specify one of: -groupdoc, -modeldoc, -entitydoc or -entitylist.");
                    exit(1);
                }
                break;

            case ArgumentListFormat:
                if ( listFormat != nil )
                    ErrPrintf(@"Warning: Already specified the list format, using first version.");
                else {
                    NSString *format = [NSString stringWithString:[argEnum nextObject]];

                    if ( [format length] == 0 )
                        ErrPrintf(@"Warning: List format must contain %%e at a minimum");
                    else {
                        NSRange modelPos = [format rangeOfString:@"%m"];
                        NSRange entityPos = [format rangeOfString:@"%e"];
                        NSCharacterSet *percentSet = [NSCharacterSet characterSetWithCharactersInString:@"%"];
                        
                        if ( entityPos.length == 0 )
                            ErrPrintf(@"Warning: Must specify %%e at a minimum.");
                        else if ( modelPos.length > 0 ) {
                            if ( [format numOfCharactersFromSet:percentSet] > 2 )
                                ErrPrintf(@"Warning: Too many %%'s in list format, using default.");
                            else if ( NSMaxRange(entityPos) < NSMaxRange(modelPos) ) {
                                listFormatOrder = EntityFirst;
                                listFormat = format;
                            }
                            else {
                                listFormatOrder = ModelFirst;
                                listFormat = format;
                            }
                        }
                        else if ( [format numOfCharactersFromSet:percentSet] > 1 )
                            ErrPrintf(@"Warning: Too many %%'s in list format, using default.");
                        else {
                            listFormatOrder = EntityFirst;
                            listFormat = format;
                        }
                    }
                }
                break;

            case ArgumentListSeparator:
                if ( listSeparator != nil )
                    ErrPrintf(@"Warning: Already specified the list separator, using first version.");
                else
                    listSeparator = [NSString stringWithString:[argEnum nextObject]];
                break;

                
            case ArgumentGenericEntityOnly:
            case ArgumentConcreteEntityOnly:
                if ( listEntityTypes == ArgumentUnknownOption )
                    listEntityTypes = argType;
                else {
                    ErrPrintf(@"Error: Can only specify one of: -concreteonly or -genericonly.");
                    exit(1);
                }
                break;



            case ArgumentDefine:
            {
                NSString *key = [currArg substringFromIndex:8];
                NSString *value = [argEnum nextObject];

                if ([key length] > 0 && value)
                    [userVariables setObject:value forKey:key];
            }
                break;


            case ArgumentUnknownOption:
                if ( [[currArg pathExtension] hasPrefix:@"eomodel"] )
                    [self addModelFromPath:currArg shouldProcess:YES];
                else if ( [currArg containsString:@"*"] || [currArg containsString:@"?"] )
                    [entityWildcards addObject:currArg];
                else
                    [entityNames addObject:currArg];
                break;

            case ArgumentInvalidOption:
            default:
                ErrPrintf(@"Invalid option: '%@'", currArg);
                [self displayHelpAndExit:1];
                break;
        }
    }

    if ( docMode == ArgumentUnknownOption ) {
        ErrPrintf(@"Must specify a document mode.  Use -help to see help.");
    }

    if ( [models count] == 0 ) {
        ErrPrintf(@"Unspecified EOModel file.  Use -help to see help.");
        exit(1);
    }

    if ( ![[NSFileManager defaultManager] directoryExistsAtPath:docDestination] ) {
        ErrPrintf(@"Destination directory (%@) does not exist.", docDestination);
        exit(1);
    }

    switch ( docMode ) {
        case ArgumentGroupDoc:
            if ( docTemplate == nil ) docTemplate = (javaMode) ? DefaultJavaGroupTemplate : DefaultObjCGroupTemplate;
            if ( docFile == nil ) docFile = DefaultGroupFile;
                break;

        case ArgumentModelDoc:
            if ( docTemplate == nil ) docTemplate = (javaMode) ? DefaultJavaModelTemplate : DefaultObjCModelTemplate;
            if ( docExtension == nil ) docExtension = DefaultModelExt;
                break;

        case ArgumentEntityDoc:
            if ( docTemplate == nil ) docTemplate = (javaMode) ? DefaultJavaEntityTemplate : DefaultObjCEntityTemplate;
            if ( docExtension == nil ) docExtension = DefaultEntityExt;
                break;

        case ArgumentEntityList:
            if ( listFormat == nil ) {
                listFormat = @"%e";
                listFormatOrder = EntityFirst;
            }
            listFormat = [listFormat stringByReplacingEveryOccurrenceOfString:@"%e" withString:@"%@"];
            listFormat = [listFormat stringByReplacingEveryOccurrenceOfString:@"%m" withString:@"%@"];
            [listFormat retain];

            if ( [listFormat hasSuffix:@"\n"] )
                listSeparator = nil;
            else if ( listSeparator == nil )
                listSeparator = @"\n";
            [listSeparator retain];
            break;
    }

    [self addDefaultSearchPaths];
}


/*** GENERATION ***/
- (NSString *)filenameForEntity:(EOEntity *)entity
{
    if ( docExtension == nil )
        return [docDestination stringByAppendingPathComponent:[entity name]];
    else
        return [docDestination stringByAppendingPathComponent:[[entity name] stringByAppendingPathExtension:docExtension]];
}

- (MiscMergeTemplate *)templateForName:(NSString *)name
{
    MiscMergeTemplate *template = nil;
    NSString *stdinput = nil;
    NSString *filename = nil;

    if ( name == nil || [name isEqualToString:@"-"] ) {
        NSFileHandle *stdinfile = [NSFileHandle fileHandleWithStandardInput];
        NSData *stdindata = [stdinfile readDataToEndOfFile];

        if ( [stdindata length] > 0 )
            stdinput = [[[NSString alloc] initWithData:stdindata encoding:NSUTF8StringEncoding] autorelease];
    }
    else {
        NSFileManager *fileManager = [NSFileManager defaultManager];

        if ([fileManager regularFileExistsAtPath:name])
            filename = name;
        else
            filename = [fileManager findFile:name inSearchPath:searchPaths];

        if (filename == nil) {
            ErrPrintf(@"Could not find template file '%@' in search path", name);
            exit(1);
        }
    }

    if ( filename != nil || stdinput != nil ) {
        if (verboseMode)
            ErrPrintf(@"Using template at path %@", filename);

        template = [[[MiscMergeTemplate alloc] init] autorelease];

        if ( verboseMode)
            ErrPrintf(@"Using delimiters '%@' and '%@'", delimStart, delimEnd);

        [template setStartDelimiter:delimStart endDelimiter:delimEnd];

        if ( filename != nil )
            [template parseContentsOfFile:filename];
        else if ( stdinput != nil )
            [template parseString:stdinput];
    }

    return template;
}

- (MiscMergeEngine *)engineWithTemplate:(MiscMergeTemplate *)template
{
    MiscMergeEngine *engine = [[[MiscMergeEngine alloc] initWithTemplate:template] autorelease];

    if ( [userVariables count] ) {
        NSEnumerator *userVariableEnum = [userVariables keyEnumerator];
        NSString *userKey;

        while (userKey = [userVariableEnum nextObject])
            [engine setEngineValue:[userVariables objectForKey:userKey] forKey:userKey];
    }

    return engine;
}

- (void)writeText:(NSString *)newVersion toFile:(NSString *)outputFilename
{
    if ( outputFilename == nil ) {
        ErrPrintf(@"%@", newVersion);
    }
    else {
        NSFileManager *fileManager = [NSFileManager defaultManager];
        NSString *oldVersion = [NSString stringWithContentsOfFile:outputFilename];

        if (oldVersion != nil && [oldVersion isEqualToString:newVersion]) {
            if (verboseMode)
                ErrPrintf(@"Output file %@ is identical to generated file; not written", outputFilename);
        }
        else if ( !forceOverwrites &&
                  [fileManager fileExistsAtPath:outputFilename] &&
                  ![fileManager isWritableFileAtPath:outputFilename] ) {
            ErrPrintf(@"ERROR: Could not overwrite read-only file '%@'", outputFilename);
        }
        else {
            NSData *fileData = [newVersion dataUsingEncoding:[NSString defaultCStringEncoding]];
            if (![fileData writeToFile:outputFilename atomically:YES]) {
                ErrPrintf(@"ERROR: Could not write file '%@'", outputFilename);
            }
            else {
                ErrPrintf(@"Generated %@\n", outputFilename);
            }
        }
    }
}


- (void)generate
{
    switch ( docMode ) {
        case ArgumentGroupDoc:
            [self generateGroupDoc];
            break;

        case ArgumentModelDoc:
            [self generateModelDoc];
            break;

        case ArgumentEntityDoc:
            [self generateEntityDoc];
            break;

        case ArgumentEntityList:
            [self generateEntityList];
            break;
    }
}


- (void)generateGroupDoc
{
    MiscMergeEngine   *engine;
    NSString          *outputFilename = nil;

    if (verboseMode)
        ErrPrintf(@"Generating model group documentation %@ from template %@", docFile, docTemplate);

    if ( ![docFile isEqualToString:@"-"] )
        outputFilename = [docDestination stringByAppendingPathComponent:docFile];

    engine = [self engineWithTemplate:[self templateForName:docTemplate]];
    [self writeText:[engine executeWithObject:[NSDictionary dictionaryWithObject:models forKey:@"models"] sender:nil]
             toFile:outputFilename];
}


- (void)generateModelDoc
{
    MiscMergeTemplate *template = [self templateForName:docTemplate];
    NSEnumerator *modelEnum = [models objectEnumerator];
    EOModel *model;

    while ( model = [modelEnum nextObject] ) {
        // if we were asked to document the model, then call this routine
        MiscMergeEngine   *engine;
        NSString          *outputFilename = nil;

        if ( docFile == nil || ![docFile isEqualToString:@"-"] )
            if ( ([models count] == 1) && (docFile != nil) )
                // Creates a file of the form 'docDestination/docFile' if only one model and docfile is defined
                outputFilename = [docDestination stringByAppendingPathComponent:docFile];
            else if ( indexFile )
                // Creates a file of the form 'docDestination/ModelName/docFile'
                outputFilename = [[docDestination stringByAppendingPathComponent:[model name]] stringByAppendingPathComponent:docFile];
            else if ( docExtension != nil )
                // Creates a file of the form 'docDestination/ModelName.docExtension'
                outputFilename = [docDestination stringByAppendingPathComponent:[[model name] stringByAppendingPathExtension:docExtension]];
            else
                // Creates a file of the form 'docDestination/ModelName'
                outputFilename = [docDestination stringByAppendingPathComponent:[model name]];

        if (verboseMode)
            ErrPrintf(@"Generating model documentation file %@ from template %@", outputFilename, docTemplate);

        engine = [[self engineWithTemplate:template] retain];
        [self writeText:[engine executeWithObject:model sender:nil] toFile:outputFilename];
        [engine release];
    }
}


- (void)generateEntity:(EOEntity *)entity withTemplate:(MiscMergeTemplate *)template
{
    NSString *entityName = [entity name];

    /*
    * Ignore EOPrototypes entities, since they're not real entities. Must allow
     * for "EOPrototypes", DB-specific ones like "EOOraclePrototypes",
     * "EOSybasePrototypes", and the "EOPrototypesToHide" one (used to hide
                                                               * default prototypes).
     */
    if (([entityName hasPrefix:@"EO"] && [entityName hasSuffix:@"Prototypes"]) ||
        ([entityName isEqualToString:@"EOPrototypesToHide"])) {
        if (verboseMode)
            ErrPrintf(@"Ignoring EOPrototypes entity '%@'", entityName);
        return;
    }
    else
    {
        MiscMergeEngine   *engine;
        NSString          *outputFilename = nil;

        outputFilename = [self filenameForEntity:entity];

        if (verboseMode)
            ErrPrintf(@"Generating class file %@ from template %@", outputFilename, docTemplate);

        engine = [[self engineWithTemplate:template] retain];
        [self writeText:[engine executeWithObject:entity sender:nil] toFile:outputFilename];
        [engine release];
    }
}

- (void)setupEntityNames
{
    /* Use specified entities instead of entire EOModel(s) if > 0 */
    if ([entityNames count] > 0 || [entityWildcards count] > 0) {
        int i;

        /* Add wildcards to entity name list, but only use list of chosen models */
        for ( i = 0; i < [entityWildcards count]; i++) {
            int j;

            for ( j = 0; j < [models count]; j++) {
                NSString *wildcard = [entityWildcards objectAtIndex:i];
                EOModel  *model = [models objectAtIndex:j];
                [entityNames addObjectsFromArray:[model entityNamesMatchingWildcard:wildcard]];
            }
        }

        if ([entityNames count] == 0) {
            ErrPrintf(@"No entity names matched wildcards %@", entityWildcards);
        }
    }
    else {
        NSEnumerator *modelEnum = [models objectEnumerator];
        EOModel *model;

        while ( model = [modelEnum nextObject] ) {
            [entityNames addObjectsFromArray:[model entityNames]];
        }
    }
}

- (void)generateEntityDoc
{
    [self setupEntityNames];

    if ( [entityNames count] > 0 ) {
        MiscMergeTemplate *template = [self templateForName:docTemplate];
        EOModelGroup *modelGroup = [EOModelGroup defaultGroup];
        NSEnumerator *enumerator = [entityNames objectEnumerator];
        NSString *entityName;

        enumerator = [entityNames objectEnumerator];
        while (entityName = [enumerator nextObject]) {
            EOEntity *entity = [modelGroup entityNamed:entityName];

            if (entity == nil) {
                ErrPrintf(@"Entity %@ not found in model, skipping.", entityName);
            }
            else if ( listEntityTypes == ArgumentUnknownOption ) {
                [self generateEntity:entity withTemplate:template];
            }
            else if ( listEntityTypes == ArgumentGenericEntityOnly ) {
                if ( [[entity classNameWithoutPackage] isEqualToString:@"EOGenericRecord"] )
                    [self generateEntity:entity withTemplate:template];
            }
            else if ( listEntityTypes == ArgumentConcreteEntityOnly ) {
                if ( ![[entity classNameWithoutPackage] isEqualToString:@"EOGenericRecord"] )
                    [self generateEntity:entity withTemplate:template];
            }
        }
    }
}

- (void)generateEntityList
{
    [self setupEntityNames];

    if ( [entityNames count] > 0 ) {
        EOModelGroup *modelGroup = [EOModelGroup defaultGroup];
        int i;

        for ( i = 0; i < [entityNames count]; i++ ) {
            NSString *entityName = [entityNames objectAtIndex:i];
            EOEntity *entity = [modelGroup entityNamed:entityName];

            if ( i > 0 && listSeparator != nil )
                Printf(listSeparator);
            
            if (entity == nil) {
                ErrPrintf(@"Entity %@ not found in model, skipping.", entityName);
            }
            else if ( listEntityTypes == ArgumentUnknownOption ) {
                if ( listFormatOrder == EntityFirst )
                    Printf(listFormat, [entity name], [[entity model] name]);
                else
                    Printf(listFormat, [[entity model] name], [entity name]);
            }
            else if ( listEntityTypes == ArgumentGenericEntityOnly ) {
                if ( [[entity classNameWithoutPackage] isEqualToString:@"EOGenericRecord"] )
                    if ( listFormatOrder == EntityFirst )
                        Printf(listFormat, [entity name], [[entity model] name]);
                    else
                        Printf(listFormat, [[entity model] name], [entity name]);
            }
            else if ( listEntityTypes == ArgumentConcreteEntityOnly ) {
                if ( ![[entity classNameWithoutPackage] isEqualToString:@"EOGenericRecord"] )
                    if ( listFormatOrder == EntityFirst )
                        Printf(listFormat, [entity name], [[entity model] name]);
                    else
                        Printf(listFormat, [[entity model] name], [entity name]);
            }
        }

        if ( listSeparator != nil )
            Printf(@"\n");
    }
}

@end
