//
//    MiscMergeCondition.m 
//      Written by Carl Lindberg Copyright 2000 by Carl Lindberg.
//                       All rights reserved.
//        This notice may not be removed from this source code.
//
//	This header is included in the MiscKit by permission from the author
//	and its use is governed by the MiscKit license, found in the file
//	"License.rtf" in the MiscKit distribution.  Please refer to that file
//	for a list of all applicable permissions and restrictions.
//	
//  This file formatted with 4 spaces per tabstop.

#import "MiscMergeCondition.h"
#import <Foundation/NSString.h>
#import <Foundation/NSArray.h>
#import <Foundation/NSCharacterSet.h>
#import <Foundation/NSException.h>
#import <Foundation/NSValue.h>
#import <Foundation/NSScanner.h>
#import <Foundation/NSCharacterSet.h>
#import <MiscFoundation.subproj/NSString+MiscAdditions.h>

@implementation MiscMergeCondition

- (BOOL)evaluateWithEngine:(MiscMergeEngine *)anEngine
{
    [NSException raise:NSInternalInconsistencyException format:@"%@ must implement -%@!",
        NSStringFromClass([self class]), NSStringFromSelector(_cmd)];
    return NO;
}

- (NSString *)descriptionWithGroupParentheses:(BOOL)showParens
{
    return [super description];
}

- (NSString *)description
{
    return [self descriptionWithGroupParentheses:NO];
}

@end


static NSString *_GroupDescription(NSArray *conditions, NSString *separator, BOOL showParens)
{
    NSString *baseString;
    int      i, count = [conditions count];
    NSMutableArray *descriptionArray = [NSMutableArray arrayWithCapacity:count];

    for (i=0; i<count; i++)
    {
        NSString *desc = [[conditions objectAtIndex:i] descriptionWithGroupParentheses:YES];
        [descriptionArray addObject:desc? desc:@""];
    }

    baseString = [descriptionArray componentsJoinedByString:separator];

    if ( showParens )
        return [NSString stringWithFormat:@"( %@ )", baseString];
    else
        return baseString;
}

@implementation MiscMergeAndCondition

- (id)initWithConditionArray:(NSArray *)conditions
{
    [super init];
    conditionArray = [[NSArray alloc] initWithArray:conditions];
    return self;
}

- (void)dealloc
{
    [conditionArray release];
    [super dealloc];
}

- (NSArray *)conditions
{
    return conditionArray;
}

- (BOOL)evaluateWithEngine:(MiscMergeEngine *)anEngine
{
    int i, count = [conditionArray count];

    if (count == 0) return NO; //error?

    for (i=0; i<count; i++)
    {
        if (![[conditionArray objectAtIndex:i] evaluateWithEngine:anEngine])
            return NO;
    }

    return YES;
}

- (NSString *)descriptionWithGroupParentheses:(BOOL)showParens
{
    return _GroupDescription(conditionArray, @" && ", showParens);
}

@end

@implementation MiscMergeOrCondition

- (id)initWithConditionArray:(NSArray *)conditions
{
    [super init];
    conditionArray = [[NSArray alloc] initWithArray:conditions];
    return self;
}

- (void)dealloc
{
    [conditionArray release];
    [super dealloc];
}

- (NSArray *)conditions
{
    return conditionArray;
}

- (BOOL)evaluateWithEngine:(MiscMergeEngine *)anEngine
{
    int i, count = [conditionArray count];

    if (count == 0) return NO; //error?

    for (i=0; i<count; i++)
    {
        if ([[conditionArray objectAtIndex:i] evaluateWithEngine:anEngine])
            return YES;
    }

    return NO;
}

- (NSString *)descriptionWithGroupParentheses:(BOOL)showParens
{
    return _GroupDescription(conditionArray, @" || ", showParens);
}

@end

@implementation MiscMergeNotCondition

- (id)initWithCondition:(MiscMergeCondition *)aCondition
{
    [super init];
    condition = [aCondition retain];
    return self;
}

- (void)dealloc
{
    [condition release];
    [super dealloc];
}

- (MiscMergeCondition *)condition
{
    return condition;
}

- (BOOL)evaluateWithEngine:(MiscMergeEngine *)anEngine
{
    return ![condition evaluateWithEngine:anEngine];
}

- (NSString *)descriptionWithGroupParentheses:(BOOL)showParens
{
    return [NSString stringWithFormat:@"!(%@)", [condition descriptionWithGroupParentheses:NO]];
}

@end


@implementation MiscMergeSimpleCondition

- (id)initWithLeftField:(NSString *)leftArg
{
    return [self initWithLeftField:leftArg operator:MiscMergeOperatorNone rightField:nil];
}

- (id)initWithLeftField:(NSString *)leftArg
               operator:(MiscMergeConditionalOperator)anOperator
             rightField:(NSString *)rightArg
{
    [super init];
    leftField = [leftArg retain];
    operator = anOperator;
    rightField = [rightArg retain];
    return self;
}

- (void)dealloc
{
    [leftField release];
    [rightField release];
    [super dealloc];
}

- (NSString *)leftField {
    return leftField;
}
- (NSString *)rightField {
    return rightField;
}
- (MiscMergeConditionalOperator)operator {
    return operator;
}
- (NSString *)operatorString
{
    switch (operator) {
        case MiscMergeOperatorEqual:              return @"==";
        case MiscMergeOperatorNotEqual:           return @"!=";
        case MiscMergeOperatorLessThanOrEqual:    return @"<=";
        case MiscMergeOperatorGreaterThanOrEqual: return @">=";
        case MiscMergeOperatorLessThan:           return @"<";
        case MiscMergeOperatorGreaterThan:        return @">";
        case MiscMergeOperatorNone:
        default:
            return nil;
    }
}

static BOOL isNumber(id anObject)
{
    if ([anObject isKindOfClass:[NSNumber class]]) return YES;
    if ([anObject isKindOfClass:[NSString class]])
    {
        NSScanner *scanner = [NSScanner scannerWithString:anObject];
        float  floatValue;
        return [scanner scanFloat:&floatValue];
    }

    return NO;
}

static BOOL isString(id anObject)
{
    return [anObject isKindOfClass:[NSString class]];
}

static NSComparisonResult compareFloats(float num1, float num2)
{
    if (num1 < num2) return NSOrderedAscending;
    if (num1 > num2) return NSOrderedDescending;
    return NSOrderedSame;
}

static Class CommonAnscestorClass(id obj1, id obj2)
{
    Class currSuperclass = [obj1 class];

    for (; currSuperclass != nil; currSuperclass = [currSuperclass superclass])
    {
        Class currTestClass = [obj2 class];
        for (; currTestClass != nil; currTestClass = [currTestClass superclass])
        {
            if (currTestClass == currSuperclass)
                return currSuperclass;
        }
    }

    return nil;
}

- (BOOL)evaluateWithEngine:(MiscMergeEngine *)anEngine
{
    NSString *upperLeft = [leftField uppercaseString];
    id       leftOperand;
    id       rightOperand;
    NSComparisonResult comparison;

    // special evaluators for first argument
    if ( [upperLeft isEqualToString:@"--NONE--"] )      return NO;
    else if ( [upperLeft isEqualToString:@"--NO--"] )   return NO;
    else if ( [upperLeft isEqualToString:@"--ALL--"] )  return YES;
    else if ( [upperLeft isEqualToString:@"--YES--"] )  return YES;

    // turn the operands into their field values; literals will
    // come through unchanged...hopefully. Hm.
    leftOperand = [anEngine valueForField:leftField];

    if (operator == MiscMergeOperatorNone)
    {
        if (isNumber(leftOperand)) return ([leftOperand floatValue]) ? YES : NO;
        if (isString(leftOperand)) return ([leftOperand length] > 0);

        // return YES if the value is non-nil.
        return ([leftField length] > 0 && leftOperand != nil);
    }
    else
    {
        rightOperand = [anEngine valueForField:rightField];

        if (leftOperand == nil && rightOperand == nil) {
            comparison = NSOrderedSame;
        }
        else if (leftOperand == nil && rightOperand != nil) {
            comparison = NSOrderedAscending;
        }
        else if (leftOperand != nil && rightOperand == nil) {
            comparison = NSOrderedDescending;
        }
        else if (isNumber(leftOperand) && isNumber(rightOperand)) {
            comparison = compareFloats([leftOperand floatValue], [rightOperand floatValue]);
        }
        else if ([leftOperand isEqual:rightOperand]) {
            comparison = NSOrderedSame;
        }
        else if ([CommonAnscestorClass(leftOperand, rightOperand)
                         instancesRespondToSelector:@selector(compare:)])
        {
            comparison = [(NSString*)leftOperand compare:rightOperand];
        }
        else { //??
            comparison = [[leftOperand description] compare:[rightOperand description]];
        }

        /*
         * now that we have comparison results, turn them into a YES/NO
         * depending upon the chosen operator.
         */
        switch (operator) {
            case MiscMergeOperatorEqual:              return (comparison == NSOrderedSame);
            case MiscMergeOperatorNotEqual:           return (comparison != NSOrderedSame);
            case MiscMergeOperatorLessThanOrEqual:    return (comparison != NSOrderedDescending);
            case MiscMergeOperatorGreaterThanOrEqual: return (comparison != NSOrderedAscending);
            case MiscMergeOperatorLessThan:           return (comparison == NSOrderedAscending);
            case MiscMergeOperatorGreaterThan:        return (comparison == NSOrderedDescending);
            case MiscMergeOperatorNone:               return NO; // handled above
        }

        return NO;
    }
}


/* Grumble ... how to do this better? */
static NSString *_QuotedDescription(NSString *string)
{
    NSString *descString;

    if (string == nil) return nil;
    descString = [[NSArray arrayWithObject:string] description];
    descString = [descString substringWithRange:NSMakeRange(1, [descString length]-2)];
    return [descString stringByTrimmingWhitespace];
}

- (NSString *)descriptionWithGroupParentheses:(BOOL)showParens
{
    if ( operator == MiscMergeOperatorNone )
    {
        return _QuotedDescription(leftField);
    }
    else
    {
        NSString *opDescription = [self operatorString];

        if (opDescription == nil)
            opDescription = @"<unknownop>";

        return [NSString stringWithFormat:@"%@ %@ %@",
            _QuotedDescription(leftField), opDescription, _QuotedDescription(rightField)];
    }
}

@end


/* This feels so backwards... */
@interface __MiscMergeConditionParsingDummyCommand : MiscMergeCommand
@end
@implementation __MiscMergeConditionParsingDummyCommand
- (void)error_closeparens {  NSLog(@"Error closeParens"); }
@end

@implementation MiscMergeCondition (ParsingExtras)

+ (MiscMergeCondition *)conditionWithString:(NSString *)string
{
    static NSCharacterSet *spaceSet = nil;
    NSScanner *scanner;
    MiscMergeCommand *dummyCommand;

    if (spaceSet == nil)
      spaceSet = [[NSCharacterSet characterSetWithCharactersInString:@" \n\t\r\v\f"] retain];

    scanner = [NSScanner scannerWithString:string];
    [scanner setCharactersToBeSkipped:spaceSet];

    dummyCommand = [[[__MiscMergeConditionParsingDummyCommand alloc] init] autorelease];
    return [dummyCommand getConditionFromScanner:scanner];
}

@end
