//
//  _MiscMergeLoopCommand.m
//            Written by Don Yacktman and Carl Lindberg
//        Copyright 1998 by Don Yacktman and Carl Lindberg.
//                     All rights reserved.
//      This notice may not be removed from this source code.
//
//    This header is included in the MiscKit by permission from the author
//    and its use is governed by the MiscKit license, found in the file
//    "License.rtf" in the MiscKit distribution.  Please refer to that file
//    for a list of all applicable permissions and restrictions.
//    

#import "_MiscMergeLoopCommand.h"
#import <Foundation/NSUtilities.h>
#import <Foundation/NSArray.h>
#import <Foundation/NSString.h>
#import <Foundation/NSDictionary.h>
#import <Foundation/NSCharacterSet.h>
#import "MiscMergeCommandBlock.h"

@implementation _MiscMergeLoopCommand

- init
{
    [super init];
    commandBlock = [[MiscMergeCommandBlock alloc] initWithOwner:self];
    return self;
}

- (void)dealloc
{
    [indexName release];
    [loopName release];
    [commandBlock release];
    [startKey release];
    [stopKey release];
    [stepKey release];
    [super dealloc];
}

static inline BOOL IsNumber(NSString *string)
{
    NSCharacterSet *digitSet = [NSCharacterSet decimalDigitCharacterSet];

    return [string length] == 0 ||
           [digitSet characterIsMember:[string characterAtIndex:0]];
}

- (NSString *)validateStep:(int)step
{
    if (step == 0)
    {
        return [NSString stringWithFormat:@"%@: Loop %@ is infinite (no step value).",
                      [self class], loopName];
    }

    return nil;
}

- (NSString *)validateStart:(int)start stop:(int)stop step:(int)step
{
    NSString *error = nil;

    if (step == 0)
    {
        step = (start <= stop)? 1 : -1;
        error = [NSString stringWithFormat:
                      @"%@: Loop %@ is infinite (no step value).  Setting to %d.\n",
                      [self class], loopName, step];
    }

    if (((start < stop) && (step < 0)) || ((start > stop) && (step > 0)))
    {
        return [NSString stringWithFormat:
            @"%@%@: Loop %@ is probably longer than you want. (start,end,step) = (%d,%d,%d).",
            error?error:@"", [self class], loopName, start, stop, step];
    }

    return error;
}

- (BOOL)parseFromScanner:(NSScanner *)aScanner template:(MiscMergeTemplate *)template
{
    [self eatKeyWord:@"loop" fromScanner:aScanner isOptional:NO];
    
    indexName  = [[self getArgumentStringFromScanner:aScanner toEnd:NO] retain];
    startKey   = [[self getArgumentStringFromScanner:aScanner toEnd:NO] retain];
    stopKey    = [[self getArgumentStringFromScanner:aScanner toEnd:NO] retain];
    stepKey    = [[self getArgumentStringFromScanner:aScanner toEnd:NO] retain];
    loopName   = [[self getArgumentStringFromScanner:aScanner toEnd:NO] retain];

    [template pushCommandBlock:commandBlock];

    /* Validate if values are hardcoded */
    if (IsNumber(startKey) && IsNumber(stopKey) && IsNumber(stepKey))
    {
        NSString *error;
        error = [self validateStart:[startKey intValue] stop:[stopKey intValue] step:[stepKey intValue]];
        if (error) [template reportParseError:@"%@", error];
        warned = YES;
    }
    else if (IsNumber(stepKey))
    {
        if ([stepKey intValue] == 0)
            [template reportParseError:@"%@: Loop %@ is infinite (no step value).",
                      [self class], loopName];
    }

    return YES;
}

- (NSString *)loopName
{
    return loopName;
}

- (void)handleEndLoopInTemplate:(MiscMergeTemplate *)template
{
    [template popCommandBlock:commandBlock];
}

- (void)executeForMerge:(MiscMergeEngine *)aMerger
{
    NSMutableDictionary *loopContext = [NSMutableDictionary dictionary];
    int theIndex;
    int start = [[aMerger valueForField:startKey] intValue];
    int stop  = [[aMerger valueForField:stopKey] intValue];
    int step  = [[aMerger valueForField:stepKey] intValue];

    if (!warned)
    {
        NSString *error = [self validateStart:start stop:stop step:step];
        if (error) {
            NSLog(@"%@", error);
            warned = YES;
        }
    }
    
    if (step == 0)
        step = (start <= stop)? 1 : -1;

    [aMerger addContextObject:loopContext];
    for (theIndex = start; (step > 0)? theIndex <= stop : theIndex >= stop; theIndex += step)
    {
        [loopContext setObject:[NSString stringWithFormat:@"%d", theIndex] forKey:indexName];
        [aMerger executeCommandBlock:commandBlock];
    }
    [aMerger removeContextObject:loopContext];
}

@end

